/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * dos2unix '\n' convertor 0.7.1
 *   based on Unix2Dos 0.9.0 by Peter Hanecak (made 19.2.1997)
 * Copyright 1997,.. by Peter Hanecak <hanecak@megaloman.sk>.
 * All rights reserved.
 *
 * Some portions by Rob Ginda <rginda@netscape.com> (07.2.2001)
 *
 * dos2unix filters reading input from stdin and writing output to stdout.
 * Without arguments it reverts the format (e.i. if source is in UNIX format,
 * output is in DOS format and vice versa).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/stat.h>
#include <unistd.h>

/* conversions */
#define CT_AUTO     0
#define CT_UNIX2DOS 1
#define CT_DOS2UNIX 2
#define CT_SKIP     3
#define CT_COPY     4

/* input data formats (bit mask) */
#define FT_DOS    1
#define FT_UNIX   2
#define FT_MIXED  3
#define FT_BINARY 4

/* input data flags (bit mask) */
#define FF_ROUGER 1

static struct option long_options[] = {
    { "auto", no_argument, NULL, 'A' },
    { "u2d", no_argument, NULL, 'D' },
    { "d2u", no_argument, NULL, 'U' },
    { "skipbin", no_argument, NULL, 'b' },
    { "help", no_argument, NULL, 'h' },
    { "test", no_argument, NULL, 't' },
    { "verbose", no_argument, NULL, 'v' },
    { 0, 0, 0, 0 }
};

char testmode = 0, verbose = 0, skipbin = 0;
char *argv0;

/* if fn is NULL then input is stdin and output is stdout */
int convert(char *fn, int convType) {
    char c;
    int fileType = 0, fileFlags = 0;
    char *tempFn = NULL;
    FILE *in = NULL, *out = NULL;
    struct stat buf;

    /* ugly little macro to clean up and return */
#define BAIL(_rval, _msg) {                                                    \
        if (in && in != stdin)                                                 \
            if (fclose(in) < 0) {                                              \
                perror(argv0); return -2;                                      \
            }                                                                  \
        if (out && out != stdout)                                              \
            if (fclose(out) < 0) {                                             \
                perror(argv0); return -2;                                      \
            }                                                                  \
        if (tempFn && remove(tempFn) < 0) {                                    \
            perror(argv0); return -2;                                          \
        }                                                                      \
        if (_rval != 0)                                                        \
            perror(_msg);                                                      \
        return _rval;                                                          \
    }

    /* open the input file, if it isn't stdin */
    if (fn != NULL) {
        if ((in = fopen(fn, "r")) == NULL)
            BAIL (-1, fn);
        if (stat(fn, &buf) != 0)
            BAIL (-1, fn);
        if (S_ISDIR(buf.st_mode)) {
            fprintf(stderr, "%s is a directory!\n", fn);
            BAIL (0, NULL);
        }
    }
    else
        in = stdin;

    /* open a temp file to store input */
    tempFn = tmpnam(NULL);
    if ((out = fopen(tempFn, "w")) == NULL)
        BAIL (-2, argv0);

    /* initial scan for filetype and fileflags
     * also store input to temp file for later conversion */
    while ((c = fgetc(in)) != EOF) {
        fputc(c, out);
        if (c == '\r') {
            c = fgetc(in);
            fputc(c, out);
            if (c != '\n')
                fileFlags |= FF_ROUGER;
            fileType |= FT_DOS;
        } else if (c == '\n') {
            fileType |= FT_UNIX;
        } else if ((c < 32 || c > 126) && c != '\t') {
            fileType |= FT_BINARY;
        }
    }

    /* if we're not working with stdin, close source (in) */
    if (fn != NULL) {
        if (fclose(in) != 0, in = NULL)
            BAIL (-2, argv0);
    }

    /* close temp file (out) */
    if (fclose(out) != 0, out = NULL)
        BAIL (-2, argv0);


    /* figure out conversion type */
    if ((fileType & FT_BINARY) != 0 && skipbin) {
            convType = CT_SKIP;
            fileType = FT_BINARY;	/* ensure we do not treat it as another format */
    }
    if (convType == CT_AUTO) {
        if ((fileType & FT_UNIX) != 0)
            convType = CT_UNIX2DOS;
        if ((fileType & FT_DOS) != 0) {
            convType = CT_DOS2UNIX;
            fileType = fileType | !FT_UNIX;	/* to ensure FT_MIXED is handled as FT_DOS */
        }
    }

    /* don't waste time on a null conversion */
    if ((fileType == FT_DOS && convType == CT_UNIX2DOS) ||
        (fileType == FT_UNIX && convType == CT_DOS2UNIX))
    {
        convType = CT_SKIP;
    }

    /* if the source is stdin, we have to reproduce input on stdout */
    if (fn == NULL && convType == CT_SKIP)
        convType = CT_COPY;


    /* verbose output: input data type */
    if (verbose) {
        fprintf (stderr, "File format '%s': ", fn);
        if ((fileType & FT_BINARY) != 0)
            fprintf (stderr, "looks binary, ");
        if ((fileFlags & FF_ROUGER) != 0)
            fprintf (stderr, "contains rouge '\\r', ");
        if ((fileType & FT_MIXED) == FT_MIXED)
            fprintf (stderr, "mixed line endings found");
        else {
            if ((fileType & FT_DOS) != 0)
                fprintf (stderr, "dos line endings found");
            if ((fileType & FT_UNIX) != 0)
                fprintf (stderr, "unix line endings found");
        }
        fprintf (stderr, "\n");
    }

    /* verbose output: conversion type */
    if (verbose & !testmode) {
        fprintf (stderr, "Conversion type '%s': ", fn);
        switch (convType) {
            case CT_DOS2UNIX:
                fprintf (stderr, "dos -> unix");
                break;
            case CT_UNIX2DOS:
                fprintf (stderr, "unix -> dos");
                break;
            case CT_COPY:
                fprintf (stderr, "none (copying source to target)");
                break;
            case CT_SKIP:
                fprintf (stderr, "none (skipping file)");
        }
        fprintf (stderr, "\n");
    }


    /* whether we are continuing conversion */
    if ((convType == CT_SKIP) || testmode)
        BAIL (0, NULL);


    /* open input data in temp file */
    if ((in = fopen(tempFn, "r")) == NULL)
        BAIL (-2, argv0);

    /* open the output file, if it isn't stdout */
    if (fn != NULL) {
        if ((out = fopen(fn, "w")) == NULL)
            BAIL (-2, fn);
    }
    else
        out = stdout;

    /* do the actual conversion */
    switch (convType) {
        case CT_COPY:
            while ((c = fgetc(in)) != EOF)
                fputc(c, out);
            break;
        case CT_DOS2UNIX:
            while ((c = fgetc(in)) != EOF) {
                if (c == '\r')
                    continue;
                fputc(c, out);
            }
            break;
        case CT_UNIX2DOS:
            while ((c = fgetc(in)) != EOF) {
                if (c == '\n')
                    fputc('\r', out);
                fputc(c, out);
            }
    }

    BAIL (0, NULL);

#undef BAIL

}

void help() {
    fprintf(stderr,
            "usage:\t%s [--verbose|-v] [--test|-t] [--d2u|--u2d|--auto|-D|-U|-A]\n"
            "\t\t [<file name> [...]]\n"
            "\t\t--auto, -A\toutput will be set based upon autodetection\n"
            "\t\t\t\tof source format\n"
            "\t\t--d2u, -U\toutput will be in UNIX format\n"
            "\t\t--u2d, -D\toutput will be in DOS format\n"
            "\t\t--skipbin, -b\tskip binary files\n"
            "\t\t--test, -t\tdon't write any conversion results\n"
            "\t\t\t\tusefull with --verbose to just report on source type\n"
            "\t\t--verbose, -v\tprint extra information on stderr\n"
            "\n"
            "- when no options are given then input format will be automaticaly detected\n"
            "  and converted to the opposite format on output\n"
            "- when no file is given, then stdin is used as input and"
            "  stdout as output\n"
            "- binary files will be skipped automatically\n", argv0);
}

int main(int argc, char *argv[]) {
    int convType = CT_AUTO;
    int o;

    argv0 = argv[0];

    /* process parameters */
    while ((o = getopt_long(argc, argv, "ADUbhvt", long_options, NULL)) != EOF) {
        switch (o) {
            case 'A':
                convType = CT_AUTO;
                break;
            case 'D':
                convType = CT_UNIX2DOS;
                break;
            case 'U':
                convType = CT_DOS2UNIX;
                break;
            case 'b':
                skipbin = 1;
                break;
            case 'h':
                help();
                return 0;
            case 't':
                testmode = 1;
                break;
            case 'v':
                verbose = 1;
                break;
            case '?':
                help();
                return -1;
        }
    }

    if (optind < argc) {
        while(optind < argc)
            if ((o = convert(argv[optind++], convType)) != 0)
                break;
    }
    else
        o = convert(NULL, convType);

    return o;
}
